""" IQ communication module (RS232 communication with IQ Interface)

CURPORT - last opened comport (current)
"""

#!/usr/bin/env python
# Copyright (c) 2011-2013 by IQ2 Development GmbH, Neckartenzlingen 
# All rights reserved

import serial
import serial.tools.list_ports
import iqlog
from iqdata import iqdata

VERSION = '1.1.0.1'

INFO = iqlog.NO_LOG
VERBOSE = iqlog.NO_LOG
LOG_PACKAGE = "iqcomm"

AUTOPORT = 'auto'
DEF_BAUDRATE = 115200
DEF_TIMEOUT = 0.500
CURPORT = None

ERR_CH = 0x00
IOL_CH = 0x40
IQI_CH = 0xC0

IQCOMM_WRITE_IQPARAM = 0x11
IQCOMM_READ_IQPARAM = 0x12

IQPAR_IQI_OPER_MODE = 1

IQMODE_BOOT = 0
IQMODE_IOL_MASTER = 1
IQMODE_IOL_DEVICE = 2
IQMODE_IOL_MONITOR = 3
IQMODE_FAST_DOWNLOAD = 0xFD

IQErrors = {
    0x1201: "RS232 interbyte timeout",
    0x1202: "SSP interbyte timeout",
    0x1203: "checksum error",
    0x1204: "unknown knight rider command",
    0x1205: "illegal knight rider command data",
    0x1301: "bootloader error (no details)",
    0x1401: "fast download answer interbyte timeout (1 msec)",
    0x1402: "fast download answer in process",
    0x1501: "iq parameter does not exist",
    0x1502: "iq parameter access denied",
    0x1503: "iq parameter value out of range",
    0x1601: "master appl state conflict",
    0x1602: "wrong data length",
    0x1603: "no such master iq command",
    0x1701: "device appl state conflict",
    0x1702: "wrong data length",
    0x1703: "no such device iq command",
    0x1704: "ISDU internal error",
    0x1705: "ISDU does not exist",
    0x1706: "ISDU is locked",
    0x1707: "ISDU access denied",
    0x1708: "ISDU subindex is not available",
    0x1709: "ISDU value out of range",
    0x170A: "ISDU length underrun",
    0x170B: "ISDU length overrun",
    0x170C: "ISDU value underrun",
    0x170D: "ISDU value overrun",
    0x170E: "ISDU is unavaliable",
    0x170F: "ISDU is temporary unavaliable",
    0x1710: "ISDU inconsistent set",
    0x1801: "service mode common test error",
    0x1802: "no such service mode command",
    0x1803: "wrong RAM/ROM address",
    0x1804: "ZMD access error"
}


def log(type, message, time=None):
    iqlog.log(type, LOG_PACKAGE, message, time)


def getports():
    """ Get all available comports in System, including AUTOPORT """
    ports = []
    ports.extend([port for port, __, __ in serial.tools.list_ports.comports()])
    return ports


def openport(port, baudrate=DEF_BAUDRATE, timeout=DEF_TIMEOUT):
    """ Open comport, save reference to openned in global var curport """
    global CURPORT
    CURPORT = serial.Serial(port, baudrate, 8, 'N', 1, timeout)
    log(VERBOSE, 'open port ' + port)
    return CURPORT


def connect(connect_action, comport=None, baudrate=DEF_BAUDRATE, timeout=DEF_TIMEOUT):
    """ Connect iqInterface

    using connect_action function to given comport or
    try all available comports from getports(),
    if comport is not given.
    """
    if not comport:
    # auto search iqInterface
        for comport in getports():
            try:
                connect(connect_action, comport, baudrate, timeout)
            except:
                pass
            else:
                return CURPORT
        raise IQCommExc("iqInterface not found")
    else:
        # connect to concrete comport
        try:
            port = openport(comport, baudrate, timeout)
            connect_action()
        except:
            disconnect(port)
            raise IQCommExc("iqInterface is not connected to port " + comport)
        else:
            log(VERBOSE, 'iqcomm: connect to port ' + comport)
            return CURPORT


def disconnect(port=None):
    """ Disconnect iqInterface (close comport)
    use global var CURPORT, if port is not defined
    """
    global CURPORT
    if not port:
        port = CURPORT
    if port:
        log(VERBOSE, 'disconnect port ' + port.port)
    if port:
        return port.close()
    if port == CURPORT:
        CURPORT = None


def sendrecv(command, data=None, port=None, channel=IOL_CH):
    """ Send iq command with data and receive response data
    return response data or error (int)
    """
    if not data:
        data = []
    if not port:
        port = CURPORT
    if not port:
        raise IQCommExc("No iq communication")
    if len(data) > 256:
        raise IQCommExc("iqcomm: can not send %d bytes, max send length is 256"
                        % len(data))

    # pack request
    sendbytes = iqdata()
    sendbytes.append(len(data) + 2)
    sendbytes.append(channel)
    sendbytes.append(command)
    if data:
        sendbytes += data
    sendbytes[1] |= get6bitcks(sendbytes)

    log(VERBOSE, 'send message ' + sendbytes.hex)

    # send request
    if port.write(sendbytes.str) != len(sendbytes):
        raise IQCommExc('iqcomm: can not send all bytes')

    # get response
    recvbytes = iqdata(port.read(1), 'str')
    if len(recvbytes) < 1:
        raise IQCommExc('iqcomm: no response or communication timeout')
    if recvbytes[0] != 0:
        restbytes = iqdata(port.read(recvbytes[0]), 'str')
        if len(restbytes) < recvbytes[0]:
            raise IQCommExc('iqcomm: no response or communication timeout')
        recvbytes += restbytes
    else:
        raise IQCommExc('iqcomm: response has zero length')

    log(VERBOSE, 'recv message ' + recvbytes.hex)

    # check checksum
    if recvbytes[1] & 0x3F != get6bitcks(recvbytes):
        raise IQCommExc("iqcomm: response checksum error")

    # check error channel
    if (recvbytes[1] & 0xC0) == ERR_CH:
        errorcode = ((recvbytes[3] << 8) & 0xFF00) + recvbytes[4]
        errmessage = ''
        if errorcode in IQErrors:
            errmessage += IQErrors[errorcode]
        else:
            errmessage = 'unknown'
        raise IQCommExc("iqcomm: error 0x" + "{0:04x}".format(errorcode) +
                        " (" + errmessage + ")")
    # check confirm channel
    elif (recvbytes[1] & 0xC0) != channel:
        raise IQCommExc('iqcomm: channel is not confirmed within response')
    # check confirm command
    elif recvbytes[2] != (command | 0x80):
        raise IQCommExc('iqcomm: command is not confirmed within response')

    return iqdata(recvbytes[3:])


def get6bitcks(data):
    """ Get 6 bit checksum for IQ communication

    >>> get6bitcks([1,2,3,4,5,6])
    30
    """
    if not data:
        raise IQCommExc('iqcomm: can not calculate checksum of empty frame')
    # first xor all telegram characters
    bytes12 = data[0:1]
    data[0] = 0x52
    data[1] &= 0xC0
    ucStart = reduce(lambda x, b: x ^ b, data)
    data[0:1] = bytes12
    # compress 8bit result to 6 bit
    ucChkSum1 = ucStart ^ (ucStart << 4)
    ucChkSum1 ^= ucChkSum1 >> 2
    ucChkSum2 = ucStart ^ (ucStart >> 1)
    ucChkSum3 = ((ucChkSum1 & 0x30) | ((ucChkSum2 & 0x40) >> 3) |
                 ((ucChkSum2 & 0x10) >> 2) | ((ucChkSum2 & 0x04) >> 1) |
                 (ucChkSum2 & 0x01))
    return ucChkSum3


def read_iqpar(num, port=None):
    return sendrecv(IQCOMM_READ_IQPARAM, [num], port)


def write_iqpar(num, val, port=None):
    return sendrecv(IQCOMM_WRITE_IQPARAM, [num, val], port)


def service_mode_command(command, data=None, port=None):
    return sendrecv(0x7F, [command] + data, port, IQI_CH)


def read_ram(address, port=None):
    data = iqdata(address, 'uint16')
    data.reverse()
    return service_mode_command(0x02, data, port)[0]


def write_ram(address, byte, port=None):
    data = iqdata(address, 'uint16')
    data.reverse()
    data += [byte]
    service_mode_command(0x01, data, port)


def get_ram_bit(address, bit, port=None):
    byte = read_ram(address, port)
    return (byte >> bit) & 1 == 1


def set_ram_bit(address, bit, value=True, port=None):
    byte = read_ram(address, port)
    byte = byte | (1 << bit) if value else byte & ~(1 << bit)
    write_ram(address, byte, port)


def set_ext_power(is_on):
    set_ram_bit(0xFF02, 7, is_on)


class IQCommExc(Exception):
    """ IQ communication exception """
    pass


import unittest


class TestCase(unittest.TestCase):
    def runTest(self):
        with self.assertRaises(IQCommExc):
            sendrecv(1, list(range(0, 456)))
        return



def test():
    def connect_action():
        return sendrecv(0x12, [1])

    try:
        print (connect(connect_action))
    finally:
        disconnect()
    return


if __name__ == '__main__':
#test()
    import doctest

    doctest.testmod()
    unittest.main()
  