""" IQ data module """

#!/usr/bin/env python
# Copyright (c) 2011-2013 by IQ2 Development GmbH, Neckartenzlingen 
# All rights reserved

from bitarray import bitarray
from struct import pack, unpack
import unittest


class iqdata(list):
    """ IQ data class:

    It behaves as bytes list, but has also the following attributes,
    which allow to treat it as the corresponding io-link data types:

    - str
    - bool
    - bits (bitarrray)
    - uint/int (auto bytes (un)pack)
    - uint8/int8
    - uint16/int16
    - uint32/int32
    - uint64/int64

    >>> data = iqdata(0, "uint16")
    >>> print data
    00:00

    >>> data = iqdata()
    >>> print data
    none
    >>> data = iqdata('erf', 'str')
    >>> print data.str
    erf
    >>> data = iqdata(-1234, 'int16')
    >>> print data.int16
    -1234
    >>> data = iqdata(1234, 'int16')
    >>> print data.int16
    1234
    >>> print data
    04:d2
    >>> data = iqdata()
    >>> data.uint16 = 4589
    >>> print data
    11:ed

    It can be inited in format 'XX:XX:...:XX':

    >>> data = iqdata('83:61:66:ab')
    >>> data
    83:61:66:ab
    >>> data.ihex
    '0x836166ab'
    >>> data.uint32
    2204198571L
    >>> data.int32
    -2090768725
    >>> data.uint
    2204198571L
    >>> data.int
    -2090768725
    >>> data[1:3]
    61:66

    Working with bitfields:

    >>> bitfields = {'f1': (8,15), 'f2': (16,23)} # byte 3 and 2
    >>> data = iqdata('83:61:66:ab', bitfields=bitfields)
    >>> data.bitfields = bitfields # or so
    >>> data.f1.uint8
    102
    >>> data.f2 = 7
    >>> data
    83:07:66:ab

    >>> d1 = iqd('34:af')
    >>> d2 = iqd('34:af')
    >>> d1 == d2
    True
    >>> d1 != d2
    False

    >>> bdata = iqdata(True)
    >>> bdata
    ff
    >>> bdata = iqdata(False)
    >>> bdata
    00

    >>> i = iqdata(0xa4, "uint", align=1)
    >>> i
    a4

    >>> i = iqdata(0xa4, align=2)
    >>> i
    00:a4

    """

    bitfields = None

    # override get attribute
    def __getattr__(self, name):
        fmt = _getintfmt(name, bnum=len(self))
        if name == 'hex':
            return (':'.join(["{0:02x}".format(b) for b in self])
                    if self else 'none')
        if name == 'ihex':
            return ('0x' + ''.join(["{0:02x}".format(b) for b in self])
                    if self else 'none')
        elif name == 'str':
            return ''.join([chr(b) for b in self])
        elif name == 'bool':
            return True if self[0] != 0 else False
        elif name == 'bits':
            bitarr = bitarray(endian='big')
            bitarr.frombytes(self.str)
            return bitarr
        elif fmt:
            return unpack('>' + fmt, self.str)[0]
        elif self.bitfields and name in self.bitfields:
            return self.bitfld(*(self.bitfields[name]))
        else:
            return list.__getattribute__(self, name)

    # override set attribute
    def __setattr__(self, name, value):
        fmt = _getintfmt(name, value=value)
        if name == 'hex':
            if not isinstance(value, str):
                raise IQDataTypeExc("only string 'XX:XX...' "
                                    "can be assigned to iqdata.hex")
            try:
                byte_arr = ([int('0x' + sbyte, 16) for sbyte in value.split(':')]
                            if value and value != 'none' else [])
            except:
                raise IQDataTypeExc("wrong format: only string 'XX:XX...' "
                                    "can be assigned to iqdata.hex")
            else:
                list.__init__(self, byte_arr)
        elif name == 'str':
            if not isinstance(value, str):
                raise IQDataTypeExc('only string can be assigned to iqdata.str')
            list.__init__(self, [ord(c) for c in value])
        elif name == 'bool':
            if isinstance(value, int):
                value = True if value != 0 else False
            list.__init__(self, [0xFF if value else 0x00])
        elif name == 'bits':
            if not isinstance(value, bitarray):
                raise IQDataTypeExc('only bitarray can be assigned to iqdata.bits')
            self.str = value.tobytes()
        elif self.bitfields and name in self.bitfields:
            self.set_bitfld(*(self.bitfields[name]), value=value)
        elif fmt:
            if not isinstance(value, int):
                raise IQDataTypeExc('only int objects can be assigned'
                                    ' to iqdata.(u)int(bits)')
            list.__init__(self, [ord(c) for c in pack('>' + fmt, value)])
        else:
            list.__setattr__(self, name, value)

    def set_bitfld(self, start, end, value):
        """ Get bit field value

        set bit field value between bit positions start and end inclusively

        >>> data = iqdata('FF:FF')
        >>> data.set_bitfld(4, 7, 0x01C0) # only 4 (7 - 4 + 1) lsb bits of value are significant
        >>> data
        ff:0f

        >>> data.set_bitfld(3, 3, False)
        >>> data
        ff:07

        """
        bitarr = self.bits
        if isinstance(value, iqdata):
            bitfield = value.bits
        elif isinstance(value, int):
            bitfield = iqdata(value, 'int').bits
        elif isinstance(value, bool):
            bitfield = bitarray(value)
        else:
            raise IQDataTypeExc('only iqdata, int or bool objects can be assigned'
                                ' to bitfield')
        bitlength = (end - start) + 1
        bitfield = bitfield[-bitlength:]
        if bitlength > len(bitfield):
            padding = bitarray((bitlength - len(bitfield)) * [False])
            bitfield += padding
        bitarr[-(end + 1):-start] = bitfield
        self.bits = bitarr

    def bitfld(self, start, end):
        """ Get bit field value

        Get bit field value between bit positions start and end inclusively

        >>> data = iqdata('00:FF') # 0000 0000 1111 1111 (15 - 0 bit numbers)
        >>> data.bitfld(6, 8).uint8
        3
        >>> data.bitfld(4, 14).uint16
        15
        >>> data.bitfld(4, 4).bool
        True
        >>> data.bitfld(10, 10).bool
        False
        >>> data = iqd('23:ac:56:af')
        >>> data.bitfld(8, 23)
        ac:56
        >>> data.bitfld(0, 2).uint8
        7

        """
        bitlength = (end - start) + 1
        if bitlength > 64 or bitlength < 1:
            raise IQDataTypeExc('can not get bit field of length {0}, '
                                'min is 1 bit and max is 64 bit'.format(bitlength))
        padding = 0 if bitlength % 8 == 0 else 8 - (bitlength % 8)
        bitfield = bitarray(padding * [False]) # padding to 8*n bits
        bitfield += self.bits[-(end + 1):-start] if start != 0 else self.bits[-(end + 1):]
        return iqdata(bitfield.tobytes(), 'str')

    # override to string
    def __str__(self):
        return self.hex

    # override repr
    def __repr__(self):
        return self.hex

    def __getslice__(self, i, j):
        return iqdata(list.__getslice__(self, i, j))

    def __init__(self, value=None, attr=None, bitfields=None, align=None):
        self.bitfields = bitfields
        if value is None:
            list.__init__(self, [])
        elif isinstance(value, iqdata):
            list.__init__(self, value)
        elif attr:
            self.__setattr__(attr, value)
        elif isinstance(value, str):
            self.__setattr__('hex', value)
        elif isinstance(value, bool):
            self.__setattr__('bool', value)
        elif isinstance(value, int):
            self.__setattr__('int', value)
        else:
            list.__init__(self, value)
        if len(self) < align:
            list.__init__(self, [0] * (align - len(self)) + self)


def _getintfmt(name, value=None, bnum=None):
    if name == 'uint8':
        fmt = 'B'
    elif name == 'int8':
        fmt = 'b'
    elif name == 'uint16':
        fmt = 'H'
    elif name == 'int16':
        fmt = 'h'
    elif name == 'uint32':
        fmt = 'I'
    elif name == 'int32':
        fmt = 'i'
    elif name == 'uint32':
        fmt = 'L'
    elif name == 'int32':
        fmt = 'l'
    elif name == 'uint':
        if (0 <= value <= 255) or bnum == 1:
            return _getintfmt('uint8')
        elif (0 <= value <= 65535) or bnum == 2:
            return _getintfmt('uint16')
        elif (0 <= value <= 4294967295) or bnum == 4:
            return _getintfmt('uint32')
        else:
            return None
    elif name == 'int':
        if (-128 <= value < 128) or bnum == 1:
            return _getintfmt('int8')
        elif (-32768 <= value < 32767) or bnum == 2:
            return _getintfmt('int16')
        elif (-2147483648 <= value < 2147483647) or bnum == 4:
            return _getintfmt('int32')
        else:
            return None
    else:
        fmt = None
    return fmt


class iqd(iqdata):
    """ iqd can be used instead of iqdata:

    >>> data = iqd('erf', 'str')
    >>> print data.str
    erf

    """
    pass


# Exceptions:
class IQDataTypeExc(Exception):
    """ IQ Data format exception """
    pass


# Test:
class TestCase(unittest.TestCase):
    def runTest(self):
        data = iqdata()
        self.assertEqual(str(data), 'none')
        data = iqdata(3, 'uint16')
        self.assertEqual(str(data), '00:03')
        data = iqdata('73:61:66:64')
        self.assertEqual(str(data), '73:61:66:64')
        self.assertEqual(data.str, 'safd')
        data.str = 'erf'
        self.assertEqual(str(data), '65:72:66')
        data = iqdata([0x12, 0xea])
        self.assertEqual(data.uint16, 0x12ea)
        data.uint = 0x341256
        self.assertEqual(str(data), '00:34:12:56')
        data.int = -458
        self.assertEqual(str(data), 'fe:36')
        self.assertEqual(data.int, -458)
        data.uint16 = 0x45af
        data += [0x12, 0xe3]
        self.assertEqual(data, [0x45, 0xaf, 0x12, 0xe3])
        self.assertEqual(data.hex, '45:af:12:e3')
        self.assertEqual(data.str, 'E\xaf\x12\xe3')
        data.hex = '95:fc:8b:80'
        self.assertEqual(data, [0x95, 0xfc, 0x8b, 0x80])
        with self.assertRaises(IQDataTypeExc):
            data.uint8 = 'blabla'
        with self.assertRaises(IQDataTypeExc):
            data.str = 132
        with self.assertRaises(IQDataTypeExc):
            data.hex = 132
        with self.assertRaises(IQDataTypeExc):
            data.hex = 'rt:89'


if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()