""" IQ master communication module """

#!/usr/bin/env python
# Copyright (c) 2011-2013 by IQ2 Development GmbH, Neckartenzlingen 
# All rights reserved

import iqcomm
import time
import iqlog
from iqdata import iqdata, iqd
#from iqtest import TestPackage

INFO = iqlog.NO_LOG
VERBOSE = iqlog.NO_LOG
LOG_PATestPackageCKAGE = "iqmst"
LOG_PACKAGE = "iqmst"

# master iqcomm commands
MST_IQCOMM_SET_MODE = 0x21
MST_IQCOMM_GET_STATUS = 0x22
MST_IQCOMM_SET_PDOUT = 0x23
MST_IQCOMM_OD_WRITE_REQ = 0x24
MST_IQCOMM_OD_WRITE_RSP = 0x25
MST_IQCOMM_OD_READ_REQ = 0x26
MST_IQCOMM_OD_READ_RSP = 0x27
MST_IQCOMM_GET_EVENT = 0x28

# master modes/states
INACTIVE = 0
DO = 1
DI = 2
CHK_FAULT = 3  # state
# AUTO = 3  # mode
PREOPERATE = 4
OPERATE = 5

# action timeouts
SET_OPER_MOE_TIMEOUT = 0.5
OD_OPER_TIMEOUT = 1
EVENT_WAIT_TIMEOUT = 0.5
WAIT_COMLOST_TIMEOUT = 1

MODES = {INACTIVE: 'INACTIVE', DO: 'DO', DI: 'DI', CHK_FAULT: 'CHK_FAULT',
         PREOPERATE: 'PREOPERATE', OPERATE: 'OPERATE'}

STATUS_BIT_FIELDS = {'state': (0, 2), 'pdoutvalid': (3, 3), 'pdinvalid': (4, 4),
                     'event': (5, 5), 'od': (6, 6), 'pdoutset': (7, 7)}

EVENT_QUALIFIER_BIT_FIELDS = {'instance': (16 + 0, 16 + 2), 'source': (16 + 3, 16 + 3),
                              'type': (16 + 4, 16 + 5), 'mode': (16 + 6, 16 + 7)}

EVENT_INSTANCE_UNKNOWN = 0
EVENT_INSTANCE_RESERVED1 = 1
EVENT_INSTANCE_RESERVED2 = 2
EVENT_INSTANCE_RESERVED3 = 3
EVENT_INSTANCE_APPL = 4
EVENT_INSTANCES = {EVENT_INSTANCE_UNKNOWN: 'UNKNOWN', EVENT_INSTANCE_APPL: 'APPL',
                   EVENT_INSTANCE_RESERVED1: 'reserved1',
                   EVENT_INSTANCE_RESERVED2: 'reserved2',
                   EVENT_INSTANCE_RESERVED3: 'reserved3'}

EVENT_SOURCE_REMOTE = 0
EVENT_SOURCE_LOCAL = 1
EVENT_SOURCES = {EVENT_SOURCE_REMOTE: 'REMOTE', EVENT_SOURCE_LOCAL: 'LOCAL'}

EVENT_TYPE_RESERVED = 0
EVENT_TYPE_NOTIFICATION = 1
EVENT_TYPE_WARNING = 2
EVENT_TYPE_ERROR = 3
EVENT_TYPES = {EVENT_TYPE_RESERVED: 'reserved', EVENT_TYPE_NOTIFICATION: 'NOTIFICATION',
               EVENT_TYPE_WARNING: 'WARNING', EVENT_TYPE_ERROR: 'ERROR'}

EVENT_MODE_RESERVED = 0
EVENT_MODE_SINGLE_SHOT = 1
EVENT_MODE_DISAPPEARS = 2
EVENT_MODE_APPEARS = 3
EVENT_MODES = {EVENT_MODE_RESERVED: 'reserved', EVENT_MODE_SINGLE_SHOT: 'SINGLE SHOT',
               EVENT_MODE_DISAPPEARS: 'DISAPPEARS', EVENT_MODE_APPEARS: 'APPEARS'}

EVENT_CODE_COMLOST = 0xFF22

ISDU_ERR_INDEX_NOT_AVAIL = 0x8011
ISDU_ERR_SUBINDEX_NOT_AVAIL = 0x8012
ISDU_ERR_ACCESS_DENIED = 0x8023
ISDU_ERR_PAR_OUT_OF_RANGE = 0x8030
ISDU_ERR_PAR_VALUE_OVERRUN = 0x8031
ISDU_ERR_PAR_VALUE_UNDERRUN = 0x8032
ISDU_ERR_PAR_LEN_OVERRUN = 0x8033
ISDU_ERR_PAR_LEN_UNDERRUN = 0x8034
ISDU_ERR_FUNC_NOT_AVAIL = 0x8035

# ISDU errors
ODErrors = {
    0x1000: "io-link communication error",
    0x1100: "io-link ISDU error / timeout",
    0x5600: "master - ISDU checksum error",
    0x5700: "master - ISDU illegal service primitive",
    0x7E00: "index and/or subindex discrepancy in start and get ISDU services response",
    0x7F00: "on-request data write access in process",
    0x7F01: "on-request data read access in process",
    0x8000: "device application error - no details",
    ISDU_ERR_INDEX_NOT_AVAIL: "index not available",
    ISDU_ERR_SUBINDEX_NOT_AVAIL: "subindex not available",
    0x8020: "service temporarily not available",
    0x8021: "service temporarily not available - local control",
    0x8022: "service temporarily not available - Device control",
    ISDU_ERR_ACCESS_DENIED: "access denied",
    ISDU_ERR_PAR_OUT_OF_RANGE: "parameter value out of range",
    ISDU_ERR_PAR_VALUE_OVERRUN: "parameter value above limit",
    ISDU_ERR_PAR_VALUE_UNDERRUN: "parameter value below limit",
    ISDU_ERR_PAR_LEN_OVERRUN: "parameter length overrun",
    ISDU_ERR_PAR_LEN_UNDERRUN: "parameter length underrun",
    ISDU_ERR_FUNC_NOT_AVAIL: "function not available",
    0x8036: "function temporarily unavailable",
    0x8040: "invalid parameter set",
    0x8041: "inconsistent parameter set",
    0x8082: "application not ready",
    0x8100: "vendor specific no details"
}


def log(type, message, time=None):
    iqlog.log(type, LOG_PACKAGE, message, time)


def connect(comport=None, baudrate=iqcomm.DEF_BAUDRATE, timeout=iqcomm.DEF_TIMEOUT):
    """ Try to connect to iqInterface using iqcomm
    Trial connect_action is set iq operation mode to master
    """
    def set_iqi_oper_mode_mst():
        iqcomm.write_iqpar(iqcomm.IQPAR_IQI_OPER_MODE, iqcomm.IQMODE_IOL_MASTER)

    return iqcomm.connect(set_iqi_oper_mode_mst, comport, baudrate, timeout)


def disconnect(port=None):
    """ Disconnect iqInterface """
    try:
        if get_state(port) != INACTIVE:
            set_oper_mode(INACTIVE, port=port)
    except:
        pass
    return iqcomm.disconnect(port)


def get_status(port=None):
    """ Get iqInterface IO-Link Master status info
    Its fields are described in  STATUS_BIT_FIELDS
    """
    status = iqdata(iqcomm.sendrecv(MST_IQCOMM_GET_STATUS, None, port)[0], 'uint8')
    status.bitfields = STATUS_BIT_FIELDS
    return status


def get_state(port=None):
    """ Get current iqInterface IO-Link Master state """
    state = get_status(port).state.uint8
    log(VERBOSE, 'state ' + MODES[state])
    return state


def get_od_status(port=None):
    """ Get Master On-request data status
    Whether some ISDU request is pending
    """
    state = get_status(port).od.bool
    log(VERBOSE, 'od status ' + repr(state))
    return state


def get_event_status(port=None):
    """ Get Master event status
    Whether some event has come from IO-Link Device
    and is pending to read out and confirm
    """
    state = get_status(port).event.bool
    log(VERBOSE, 'event status ' + repr(state))
    return state


def get_pdin_valid(port=None):
    state = get_status(port).pdinvalid.bool
    validity = 'invalid' if state else 'valid'
    log(INFO, 'pd in is ' + validity)
    return state


def get_pdin(port=None):
    """ Get current process data input from IO-Link Device """
    pdin = iqdata(iqcomm.sendrecv(MST_IQCOMM_GET_STATUS, None, port)[1:])
    log(INFO, 'pdin ' + pdin.hex)
    return pdin


def set_pdout(pdout, validity=True, port=None):
    """ Set current process data output from IO-Link Master """
    data = iqdata(not validity)
    data += iqdata(pdout)
    iqcomm.sendrecv(MST_IQCOMM_SET_PDOUT, data, port)
    validity = 'invalid' if validity else 'valid'
    log(INFO, 'set pdout to {0} ({1})'.format(repr(pdout), validity))


def set_oper_mode(mode, state2wait=None, port=None):
    """ Try to set new IO-Link mode to Master
    If state2wait is not given, then mode is considered to be set,
    when state has changed to this mode before timeout,
    else given state2wait is awaited
    """
    log(INFO, 'set mode ' + MODES[mode])
    if not INACTIVE <= mode <= OPERATE:
        raise IQMstExc('iqmst: wrong master operation mode')
    iqcomm.sendrecv(MST_IQCOMM_SET_MODE, [mode], port)
    init_time = time.clock()
    while True:
        state = get_state(port)
        if (state2wait and state == state2wait) or state == mode:
            break
        if (time.clock() - init_time) > SET_OPER_MOE_TIMEOUT:
            raise IQMstExc('iqmst set oper mode ' + MODES[mode] + ': '
                                                                  'state has not been properly changed during timeout')
    log(INFO, MODES[state] + ' is achieved')
    return state


def get_odread_rsp(port=None):
    """ Get On-request data response, after ISDU read request """
    return iqcomm.sendrecv(MST_IQCOMM_OD_READ_RSP, None, port)


def get_odwrite_rsp(port=None):
    """ Get On-request data response, after ISDU write request """
    return iqcomm.sendrecv(MST_IQCOMM_OD_WRITE_RSP, None, port)


def odread(index, subindex=0, port=None):
    """ ISDU read request to IO-Link Device
    Returns tuple
    (response with iqdata object or None, ISDU error integer code or None).
    If error is None, then ISDU request is ok,
    else response is None and device responded with ISDU error.
    """
    log(INFO, 'ISDU read request index 0x{0:04x} subindex 0x{1:02x}'
              .format(index, subindex))
    request = iqdata(index, 'uint16')
    request.reverse()
    request.append(subindex)
    iqcomm.sendrecv(MST_IQCOMM_OD_READ_REQ, request, port)
    init_time = time.clock()
    while get_od_status(port):
        if (time.clock() - init_time) > OD_OPER_TIMEOUT:
            raise IQMstExc('iqmst: od read response wait timeout')
    response = get_odread_rsp(port)
    error = None
    if response[0] == 0:
        response.pop(0)
        log(INFO, 'ISDU read response ' + response.hex)
    else:
        error = response.uint16
        log(INFO, 'ISDU read error response ' + response.ihex)
        response = None
    return response, error


def odread_check_no_error(index, subindex=0, port=None):
    """ ISDU read request to IO-Link Device with check of no error
    Generate exception, if ISDU response contains ISDU error
    """
    response, error = odread(index, subindex, port)
    if not error is None:
        raise IQMstExc('iqmst: ISDU read request '
                       'index 0x{0:04x} subindex 0x{1:02x} failed, '
                       .format(index, subindex) + oderror_message(error))
    return response


def odread_check_error(expected_error, index, subindex=0, port=None):
    """ ISDU read request to IO-Link Device with error check
    Generate exception, if ISDU response does not contain expected ISDU error
    """
    response, error = odread(index, subindex, port)
    if error != expected_error:
        raise IQMstExc('iqmst: ISDU read error response '
                       'index 0x{0:04x} subindex 0x{1:02x} is wrong.\n'
                       'GOT: {2}\nEXPECTED {3}'
                       .format(index, subindex, oderror_message(error), oderror_message(expected_error)))
    return response


def odwrite(index, subindex=0, data=None, port=None):
    """ ISDU write request to IO-Link Device
    data can be iqdata object or list
    if device responded with ISDU error,
    returns ISDU error integer code else returns None.
    """
    log(INFO, 'ISDU write request index 0x{0:04x} subindex 0x{1:02x} data {2}'
              .format(index, subindex, data))
    data = iqdata(data)
    request = iqdata(index, 'uint16')
    request.reverse()
    request.append(subindex)
    request += data
    iqcomm.sendrecv(MST_IQCOMM_OD_WRITE_REQ, request, port)
    init_time = time.clock()
    while get_od_status(port):
        if (time.clock() - init_time) > OD_OPER_TIMEOUT:
            raise IQMstExc('iqmst: od write response wait timeout')
    response = get_odwrite_rsp(port)
    error = None
    if response[0] == 0:
        log(INFO, 'ISDU write response ok')
    else:
        error = response.uint16
        log(INFO, 'ISDU write error response ' + response.ihex)
    return error


def odwrite_check_no_error(index, subindex=0, data=None, port=None):
    """ ISDU write request to IO-Link Device with check of no error
    Generate exception, if ISDU response contains ISDU error
    """
    error = odwrite(index, subindex, data, port)
    if not error is None:
        raise IQMstExc('iqmst: ISDU write request '
                       'index 0x{0:04x} subindex 0x{1:02x} data {2} failed, '
                       .format(index, subindex, data) + oderror_message(error))


def odwrite_check_error(expected_error, index, subindex=0, data=None, port=None):
    """ ISDU write request to IO-Link Device with error check
    Generate exception, if ISDU response does not contain expected ISDU error
    """
    error = odwrite(index, subindex, data, port)
    if error != expected_error:
        raise IQMstExc('iqmst: ISDU write error response '
                       'index 0x{0:04x} subindex 0x{1:02x} data {2} is wrong.\n'
                       'GOT: {3}\nEXPECTED {4}'
                       .format(index, subindex, data, oderror_message(error), oderror_message(expected_error)))


def oderror_message(error):
    """ Compose IO-Link ISDU error message from integer error code """
    if error is None:
        return 'no ISDU error'
    else:
        description = ODErrors[error] if error in ODErrors else "unknown"
        return "ISDU error 0x{0:04x} - ".format(error) + description


def read_event(port=None):
    """ Try to wait IO-Link event from device and read it out from master
    Returns event iqdata object with fields from EVENT_QUALIFIER_BIT_FIELDS
    and integer code attribute or None, if timeout occurred.
    """
    init_time = time.clock()
    while not get_event_status(port):
        if (time.clock() - init_time) > EVENT_WAIT_TIMEOUT:
            log(INFO, 'no new event or timeout')
            return None
    event = iqcomm.sendrecv(MST_IQCOMM_GET_EVENT, None, port)
    event.bitfields = EVENT_QUALIFIER_BIT_FIELDS
    event.qualifier = event[0]
    event.code = event[1:].uint16
    log(INFO, 'new EVENT - instance: {0}, source: {1}, type: {2}, mode: {3}, code: {4}'
              .format(EVENT_INSTANCES[event.instance.uint8],
                      EVENT_SOURCES[event.source.uint8],
                      EVENT_TYPES[event.type.uint8],
                      EVENT_MODES[event.mode.uint8], event[1:].ihex))
    return event


def wait_comlost(timeout=None, port=None):
    if timeout is None:
        timeout = WAIT_COMLOST_TIMEOUT
    log(INFO, 'wait comlost event')
    init_time = time.clock()
    while read_event(port).code != EVENT_CODE_COMLOST:
        if (time.clock() - init_time) > timeout:
            raise IQMstExc('iqmst: timeout waiting comlost')


class IQMstExc(Exception):
    """ IQ master communication exception """
    pass


def test():
    index = 0x01ea
    subindex = 0
    wdata = iqd('00:00')
    try:
        connect()
        set_oper_mode(OPERATE)
        odwrite(index, subindex, wdata)
        rdata = odread(index, subindex)
        # assert wdata != rdata
        # read_event()
        # get_pdin()
        # get_pdin_valid()
        # set_pdout(1)
    finally:
        disconnect()


# class IQMstTestPackage(TestPackage):
    # def __init__(self, options=None, iol_mode=None, comport=None):
        # self.iol_mode = OPERATE if iol_mode is None else iol_mode
        # TestPackage.__init__(self, options, comport)

    # def before_run(self):
        # TestPackage.before_run(self)
        # self.port = connect(self.comport)
        # set_oper_mode(self.iol_mode, self.iol_mode, self.port)

    # def after_run(self):
        # disconnect(self.port)
        # TestPackage.after_run(self)


if __name__ == '__main__':
    test()
    #import doctest
    #doctest.testmod()
    #unittest.main()