""" IQ test procedure module """

#!/usr/bin/env python
# Copyright (c) 2011-2013 by IQ2 Development GmbH, Neckartenzlingen
# All rights reserved

from time import sleep
from datetime import datetime
import iqlog
import iqcomm
import iqmst
import traceback

INFO = iqlog.NO_LOG
VERBOSE = iqlog.NO_LOG
LOG_PACKAGE = "iqtest"

OPT_HELP = '-h'
OPT_VERBOSE = '-v'
OPT_INFO = '-i'
OPT_TRACE = '-t'
OPT_COMPORT = '-p'
OPT_TEST_NAME = '-tn'

OPTIONS = {OPT_HELP: 'help info',
           OPT_VERBOSE: 'verbose log mode  (=module1,module2 or for all modules, if no values)',
           OPT_INFO: 'info log mode (=module1,module2,.. or for all modules, if no values)',
           OPT_TRACE: 'log trace if exception occur',
           OPT_COMPORT: 'comport to connect (COM3,COM12..)',
           OPT_TEST_NAME: 'run only tests with given names (=test_name1,..)'}

MODULE_IQCOMM = 'com'
MODULE_IQMST = 'mst'
MODULE_IQDEV = 'dev'
MODULE_IQTEST = 'tst'

MODULES = {MODULE_IQCOMM: 'iq commmunication',
           MODULE_IQMST: 'iq master',
           MODULE_IQDEV: 'iq device',
           MODULE_IQTEST: 'iq testing'}


def log(type, message, time=None):
    iqlog.log(type, LOG_PACKAGE, message, time)


def log_error(message, time=None):
        log(iqlog.ERROR_LOG, message, time)


def log_verbose(message, time=None):
        log(VERBOSE, message, time)


def log_info(message, time=None):
    log(INFO, message, time)


class IQTestExc(Exception):
    pass


def check_true(exp, message=None):
    if message is None:
        message = ''
    if not exp:
        raise IQTestExc(message + ': expression is not true')


def check_equals(exp1, exp2, message=None):
    if message is None:
        message = ''
    if exp1 != exp2:
        raise IQTestExc(message + ': equality check fails value1 {0} value2 {1}'
                        .format(exp1, exp2))


def check_seq_equals(seq1, seq2, message=None):
    if message is None:
        message = ""
    if len(seq1) != len(seq2):
        raise IQTestExc(message + ': sequences have different lengths {0} and {1}'.format(len(seq1), len(seq2)))
    for i in range(0, len(seq1)):
        if seq1[i] != seq2[i]:
            raise IQTestExc(message + ': sequences differ at index {0} value1 {1} value2 {2}'
                            .format(i, seq1[i], seq2[i]))


class TestCase(object):
    """ IQ Test case base class """
    def __init__(self, options=None, port=None):
        self.skip = False
        self.port = port
        self.is_ok = False
        self.init()
        self.run_test(options)

    def run_test(self, options=None):
        print ('='*60)
        print ("TEST CASE: " + self.__class__.__name__)
        print ("DESCRIPTION: " + str(self.__doc__))
        err_message = ""
        enter = datetime.now()
        try:
            self.is_ok = self.run()
        except Exception as e:
            self.is_ok = False
            if not options is None and OPT_TRACE in options:
                traceback.print_exc()
            else:
                err_message = ': \n' + e.message + '\n'
        if self.is_ok is None:
                self.is_ok = True
        result = "Ok " if self.is_ok else "Failed" + err_message
        time = datetime.now() - enter
        print ("TEST CASE " + result + "Time: " + str(time))
        print ('='*60)

    def init(self):
        pass

    def run(self):
        print ("emtpy test case")
        return True


class TestPackage():
    """ IQ Test package base class """
    def __init__(self, options=None, comport=None):
        self.skip = False
        self.comport = comport
        self.port = None
        self.options = {}
        self.parse_options(options)
        if OPT_HELP in options:
            self.print_help_info()
            return
        module = self.__class__.__module__
        self.module = __import__(module)
        for part in module.split('.')[1:]:
            self.module = getattr(self.module, part)
        self.tests = []
        for name in dir(self.module):
            obj = getattr(self.module, name)
            is_test_case = (isinstance(obj, type) and
                            issubclass(obj, TestCase) and
                            not obj is TestCase)
            if is_test_case:
                self.tests.append(obj)
        self.pre_init()
        self.run()

    def run(self):
        print ("")
        print ('/'*60)
        print ("TEST PACKAGE: " + self.__class__.__name__)
        print ("DESCRIPTION: " + str(self.__doc__))
        print ('/'*60)
        print ("")
        ok_num = 0
        failed_num = 0
        test_cases = []
        enter = datetime.now()
        try:
            self.before_run()
            for test in self.tests:
                if OPT_TEST_NAME in self.options and not test.__name__ in self.options[OPT_TEST_NAME]:
                    continue
                test_cases.append(test(self.options, self.port))
            ok_num = sum([1 if test_case.is_ok else 0 for test_case in test_cases])
            failed_num = len(test_cases) - ok_num
            self.after_run()
        except Exception as e:
            log(iqlog.ERROR_LOG, e.message + '\n')
            if OPT_TRACE in self.options:
                traceback.print_exc()
        time = datetime.now() - enter
        print ("")
        print ('/'*60)
        print ("END TEST PACKAGE: " + self.__class__.__name__)
        print ("Run: {0}, Ok: {1}, Failed: {2}, Time: {3}".format(len(test_cases), ok_num, failed_num, time))
        print ('/'*60)
        print ("")

    def pre_init(self):
        pass

    def before_run(self):
        pass

    def after_run(self):
        pass

    def parse_options(self, options):
        global VERBOSE
        global INFO
        if options:
            for option in options:
                strs = option.split('=')
                opt_name = strs[0]
                opt_values = [] if len(strs) <= 1 else strs[1].split(',')
                self.options[opt_name] = opt_values
                if opt_name == OPT_VERBOSE:
                    if opt_values:
                        for opt_value in opt_values:
                            if opt_value == MODULE_IQTEST:
                                VERBOSE = iqlog.VERBOSE_LOG
                            elif opt_value == MODULE_IQCOMM:
                                iqcomm.VERBOSE = iqlog.VERBOSE_LOG
                            elif opt_value == MODULE_IQMST:
                                iqmst.VERBOSE = iqlog.VERBOSE_LOG
                    else:
                        VERBOSE = iqlog.VERBOSE_LOG
                        iqcomm.VERBOSE = iqlog.VERBOSE_LOG
                        iqmst.VERBOSE = iqlog.VERBOSE_LOG
                if opt_name == OPT_INFO:
                    if opt_values:
                        for opt_value in opt_values:
                            if opt_value == MODULE_IQTEST:
                                INFO = iqlog.INFO_LOG
                            elif opt_value == MODULE_IQCOMM:
                                iqcomm.INFO = iqlog.INFO_LOG
                            elif opt_value == MODULE_IQMST:
                                iqmst.INFO = iqlog.INFO_LOG
                    else:
                        INFO = iqlog.INFO_LOG
                        iqcomm.INFO = iqlog.INFO_LOG
                        iqmst.INFO = iqlog.INFO_LOG
                if opt_name == OPT_COMPORT and opt_values:
                    self.comport = opt_values[0]

    def print_help_info(self):
        print ('usage: test_file.py [option[=option_value1[,option_value2][...]]]\n')
        print ('Options:')
        for opt in OPTIONS:
            print (opt + ": " + OPTIONS[opt])
        print ('Modules:')
        for module in MODULES:
            print (module + ": " + MODULES[module])